"""Main module of HTTP Transporter.
"""

import uvicorn
import enum
import typing
from typing import Optional as Opt
import json
import urllib.parse
import http.cookies
from ...task import Task, TaskID, TaskMetadata
from ...task.result import TaskResult, JsonBody, StreamingBody
from . import _types as http_types
from .base import MIMEType, HTTPHeader, TStatus2HCode
from ...utils.main import try_convert_str
from ...utils.enum_ import dump_enum
from ..base import BaseTransporter
from ...task.main import Method, LazyParameter
from ...exceptions import BlueFirmamentException
from ..._types import _undefined

if typing.TYPE_CHECKING:
    from ...core.app import BlueFirmamentApp


TV = typing.TypeVar("TV")
class HTTPHeaders:
    """Parse ASGI headers to provide easy access to HTTP headers.

    :ivar __raw_headers: ASGI headers
    :ivar __parsed_headers:
    :ivar __headers: for create a response headers
    """

    def __init__(self,
        raw_headers: Opt[typing.Iterable[tuple[bytes, bytes]]] = None
    ):
        self.__raw_headers = raw_headers or ()
        self.__parsed_headers: dict[str, list[str] | str] = {}
        self.__headers: dict[str, str] = {}

    def _lookup_in_raw_headers(self, key: str) -> Opt[str | list[str]]:
        res: list[str] = []
        for header in self.__raw_headers:
            if key == header[0].decode('latin-1'):
                res.append(header[1].decode('latin-1'))

        if len(res) == 0:
            return None
        if len(res) == 1:
            res = res[0]
        
        self.__parsed_headers[key] = res
        return res
    
    def get(self, key: str | enum.Enum, default: TV = None) -> list[str] | str | TV:
        res = self.__parsed_headers.get(dump_enum(key), None)
        if res is None:
            res = self._lookup_in_raw_headers(dump_enum(key))
            if res is None:
                return default
        return res

    def get_as_str(self, key: str | enum.Enum, default: str = "") -> str:
        res = self.get(key, default)
        if res is None:
            return ""
        if isinstance(res, str):
            return res
        else:
            raise TypeError("Header exist but value is not a string")
        
    def get_as_list(self, key: str | enum.Enum, default: Opt[list] = None) -> list[str]:
        res = self.get(key, None)
        if isinstance(res, list):
            return res
        if res is None:
            return default or []
        else:
            return [res]
        
    def get_content_type(self) -> tuple[Opt[MIMEType], str]:
        """Get Header 'Content-Type'

        :returns: a tuple, 0 for MIME type, 1 for charset

            If charset not set, defaults to `utf-8`.
            If MIMEType not set, defaults to None.
        """
        content_type_str = self.get_as_str(HTTPHeader.CONTENT_TYPE)
        if not content_type_str:
            return None, "utf-8"
        split = content_type_str.split(';')
        return MIMEType(split[0]), split[1].split('=')[1] if len(split) > 1 else "utf-8"

    def get_accept(self) -> tuple[MIMEType, ...]:
        """Get 'Accept' in header

        :returns: a list of MIME types that client accepts
        """
        accept_str = self.get_as_str(HTTPHeader.ACCEPT)
        if not accept_str:
            return (MIMEType.JSON,)
        return tuple(
            MIMEType(i.strip().split(';')[0])
            for i in accept_str.split(',')
        )

    def get_accept_charset(self) -> tuple[str, ...]:
        """Get 'Accept-Charset' in header

        :returns: a string of charset that client accepts
        """
        accept_charset_str = self.get_as_str(HTTPHeader.ACCEPT_CHARSET)
        if not accept_charset_str:
            return ('utf-8',)
        return tuple(
            i.strip().split(';')[0]
            for i in accept_charset_str.split(',')
        )

    @property
    def dict(self) -> dict[str, str]:
        return self.__headers

    def __setitem__(self, key: str | enum.Enum, value: str):
        self.__headers[dump_enum(key)] = value

    def set_content_type(self, mime_type: MIMEType, charset: str = "utf-8"):
        self[HTTPHeader.CONTENT_TYPE] = f"{dump_enum(mime_type)}; charset={charset}"


class HTTPBody(LazyParameter):
    """HTTP body as Task Parameter.

    Receive body bytes and parse it by MIMEType, encoding
    set in the header when needed. (lazy)
    """

    def __init__(
        self,
        receive: http_types.ASGIReceiveCallable,
        mime_type: MIMEType,
        content_encoding: str = 'utf-8',
    ):
        self.__receive = receive
        self.__mime_type = mime_type
        self.__encoding = content_encoding
        self.__parsed_body = _undefined

    async def get(self) -> typing.Any:
        if self.__parsed_body is _undefined:
            body_raw = b''
            while True:
                receive_chunk = await self.__receive()
                body_raw += receive_chunk.get('body', b'')
                if not receive_chunk.get('more_body', False):
                    break
            self.__parsed_body = self._parse_body(body_raw=body_raw)

        return self.__parsed_body

    def _parse_body(
        self, body_raw: bytes,
    ) -> typing.Union[dict[str, typing.Any], str, bytes, None]:
        """解析请求体

        将字节流根据内容类型解析为字典、字符串或字节流

        :param body_raw: 原始请求体
        :param mime_type: 内容类型
        :param content_encoding: 内容编码；默认使用 `utf-8`

        Behaviour
        ---------
        - 如果不支持的内容类型，抛出 ``ValueError`` 异常

        解析
        ^^^^^^
        - ``application/json``: 解析为字典
        - ``application/x-www-form-urlencoded``: 解析为字典
            - 如果有重复的键，使用列表存储所有值
            - 尝试将值转换为布尔、整数或浮点数，如果失败则为字符串
            - 空字符串被转换为None
        - ``text/plain``: 解析为字符串
        - ``application/octet-stream``: 返回原始字节流

        """
        if not body_raw:
            return None

        mime_type = dump_enum(self.__mime_type)
        if mime_type == MIMEType.JSON.value:
            return json.loads(body_raw.decode(self.__encoding))
        elif mime_type == MIMEType.FORM.value:
            parsed_dict = {}
            pairs = urllib.parse.parse_qsl(body_raw.decode(self.__encoding))
            for key, value in pairs:
                if key in parsed_dict:
                    if isinstance(parsed_dict[key], list):
                        parsed_dict[key].append(try_convert_str(value))
                    else:
                        parsed_dict[key] = [parsed_dict[key], try_convert_str(value)]
                else:
                    parsed_dict[key] = try_convert_str(value)
            return parsed_dict
        elif mime_type == MIMEType.TEXT.value:
            return body_raw.decode(self.__encoding)
        elif mime_type == MIMEType.BINARY.value:
            return body_raw
        else:
            raise ValueError(f'Unsupported content type: {mime_type}')


class HTTPTransporter(BaseTransporter):
    """Transporter serves HTTP/S protocol.
    """

    def __init__(
        self,
        app: "BlueFirmamentApp",
        host: str,
        port: int,
        uds: Opt[str] = None,
        name: str = "default"
    ):
        """
        :param uds: Unix domain socket. E.g /tmp/blue_firmament.sock
        """
        super().__init__(app=app, name=name)
        self.__asgi_server = uvicorn.Server(uvicorn.Config(
            app=self, host=host, port=port, uds=uds
        ))

    def start(self):
        return self.__asgi_server.serve()

    def stop(self):
        return self.__asgi_server.shutdown()

    async def __call__(self, 
        scope: http_types.Scope, 
        receive: http_types.ASGIReceiveCallable, 
        send: http_types.ASGISendCallable
    ):
        """
        `ASGI spec of applications <https://asgi.readthedocs.io/en/latest/specs/main.html#applications>`_
        """
        if scope['type'] == 'http':
            # parse headers
            headers = HTTPHeaders(scope['headers'])

            # compose task and task_result
            h_content_type = headers.get_content_type()
            task = Task(
                task_id=TaskID(
                    method=Method(scope['method']),
                    path=scope['path'],
                    separator='/'
                ),
                parameters={
                    "body": HTTPBody(
                        receive=receive,
                        mime_type=h_content_type[0],
                        content_encoding=h_content_type[1]
                    ),
                    **self.parse_query_params(scope['query_string'])
                },
                metadata=self.parse_metadata(headers)
            )
            task_result = TaskResult()

            try:
                await self._app.handle_task(
                    task=task, task_result=task_result, transporter=self
                )
            except BlueFirmamentException as e:
                task_result.status = e.task_status
                task_result.body = JsonBody(e.dump_details_to_dict())

            # send response
            try:
                res_headers = HTTPHeaders()
                if isinstance(task_result.body, JsonBody):
                    res_headers.set_content_type(MIMEType.JSON)
                elif isinstance(task_result.body, StreamingBody):
                    res_headers.set_content_type(MIMEType.EVENT_STREAM)
                    res_headers[HTTPHeader.CONNECTION] = "keep-alive"
                    res_headers[HTTPHeader.CACHE_CONTROL] = "no-cache"

                await send(http_types.HTTPResponseStartEvent(
                    type="http.response.start",
                    status=TStatus2HCode[task_result.status],
                    headers=task_result.metadata.dump_to_bytes(
                        encoding="latin-1",
                        extra=res_headers.dict
                    )
                ))

                res_body = task_result.body
                if isinstance(res_body, StreamingBody):
                    async for chunk in res_body:
                        body_ = f"data: {chunk.dump_to_str()}\n\n".encode("utf-8")
                        await send(http_types.HTTPResponseBodyEvent(
                            type="http.response.body",
                            body=body_,
                            more_body=True
                        ))

                    await send(http_types.HTTPResponseBodyEvent(
                        type="http.response.body",
                        body=b'',
                        more_body=False
                    ))
                else:
                    await send(http_types.HTTPResponseBodyEvent(
                        type="http.response.body",
                        body=res_body.dump_to_bytes(encoding="utf-8"),
                        more_body=False
                    ))

            except OSError:   # Disconnected unexpectedly
                self._logger.warning('Connection closed before all body were sent')
                await task_result.body.cleanup()
        else:
            self._logger.warning(f"Request omitted due to unsupported protocol {scope['type']}")

    @staticmethod
    def parse_metadata(headers: HTTPHeaders) -> TaskMetadata:
        # parse authorization header
        authorization = headers.get_as_str('authorization').split(" ")
        # parse cookies
        cookies = {}
        for cookie_str in headers.get_as_list('cookie', []):
            cookie = http.cookies.SimpleCookie()
            cookie.load(cookie_str)
            for name, morsel in cookie.items():
                cookies[name] = morsel.value
        return TaskMetadata(
            authorization=(authorization[0], authorization[1]) \
                if len(authorization) == 2 else None,
            trace_id=headers.get_as_str('x-trace-id'),
            client_id=headers.get_as_str('x-client-id'),
            state=cookies
        )

    @staticmethod
    def parse_query_params(
        query_bytes: bytes,
        encoding: str = 'latin-1',
    ) -> dict[str, str | int | float | bool | None | list]:
        """解析查询参数

        将查询字符串解析为字典

        :param query_bytes: 查询字符串

        :raises ValueError: unsupported value type
        
        解析
        ^^^^^^
        - 尝试将值转换为布尔、整数或浮点数，如果失败则为字符串
        - 如果有多个相同的键，使用元组存储所有值
        - 空字符串被转换为None
        
        """
        query_string = query_bytes.decode(encoding)
        parsed_dict = {}
        pairs: list[tuple[str, str]] = urllib.parse.parse_qsl(query_string)
        for key, value in pairs:
            # TODO performance optimization required
            if key in parsed_dict:
                if isinstance(parsed_dict[key], list):
                    parsed_dict[key].append(try_convert_str(value))
                else:
                    parsed_dict[key] = [parsed_dict[key], try_convert_str(value)]
            else:
                parsed_dict[key] = try_convert_str(value)
        return parsed_dict



"""Auth module
"""

__all__ = [
    "User",
    "AuthSession"
]

import abc
import datetime
import typing
from typing import Annotated as Anno, Optional as Opt, Literal as Lit

from .utils import auth_
from .utils.datetime_ import get_datetimez
from .utils.main import dump_iterable
from .data.settings.auth import get_setting as get_auth_setting
from .log import get_logger
LOGGER = get_logger(__name__)


class User(abc.ABC):

    def __init__(
        self,
        _id,
        roles: set[str]
    ):
        self.__id = _id
        self.__roles: set[str] = roles

    @property
    def id(self):
        """ID of the user.
        """
        return self.__id

    @property
    def roles(self):
        """Roles of the user.
        """
        return self.__roles

    def has_role(self, role: str) -> bool:
        return role in self.__roles


class AuthSession(abc.ABC):
    """Session of auth module.
    """

    def __init__(
        self,
        _id: str,
        access_token: str,
        user: User,
        refresh_token: Opt[str] = None,
        expire_at: Opt[datetime.datetime] = None,
    ) -> None:
        self.__id = _id
        self.__access_token = access_token
        self.__refresh_token = refresh_token
        self.__user = user
        self.__expire_at = expire_at

    @property
    def id(self):
        """Session ID"""
        return self.__id
    @property
    def user(self):
        """User of the session"""
        return self.__user
    @property
    def access_token(self):
        return self.__access_token

    def is_expired(self) -> bool:
        if self.__expire_at is None:
            return False
        return get_datetimez() > self.__expire_at

    def refresh(self) -> None:
        """Refresh the session with refresh token.

        :raise ParamsInvalid: if refresh token is not set.
        """
    
    @classmethod
    def from_token(
        cls,
        access_token: str,
        access_token_type: Lit["jwt"] | str = "jwt",
        access_token_payload: Opt[dict] = None,
        refresh_token: Opt[str] = None,
    ) -> typing.Self:
        """Init an AuthSession from an access token.

        :param access_token:
        :param access_token_type: JWT, etc.
            Defaults to "jwt".
        :param access_token_payload: When you already have the decoded the token.
        :param refresh_token:
        :return: An AuthSession instance.

        JsonWebToken
        ^^^^^^^^^^^^
        - Requires a JWT secret key and algorithm configured in
          :meth:`blue_firmament.data.setting.auth.AuthSetting`.
        - These setting are optional:
            - Allowed Audience
        - Payload claims maps to AuthSession fields, you can override
          by setting `jwt_claims_map` in AuthSetting.
            - `sid`: Session ID
            - `sub`: User ID
            - `roles`: User roles
        """
        if access_token_payload is None:
            access_token_payload = auth_.decode_token(access_token, access_token_type)

        return cls(
            access_token=access_token,
            refresh_token=refresh_token,
            **cls._normalize_token_payload(access_token_payload)
        )

    @staticmethod
    def _normalize_token_payload(payload: dict) -> dict:
        """Normalize access token payload to AuthSession fields.
        """
        getters = get_auth_setting().session_fields_getter
        return dict(
            _id=getters["session_id"](payload),
            expire_at=getters["expire_at"](payload),
            user=User(
                _id=getters["user_id"](payload),
                roles=dump_iterable(
                    set, getters["roles"](payload)
                )
            )
        )







"""Data Access Layer Module Types
"""

import typing
import enum
if typing.TYPE_CHECKING:
    from .query_components import DALQueryComponent

DALPath = typing.NewType('DALPath', typing.Tuple[str | enum.Enum | None, ...])
'''Path of DataAccessLayer

Examples
--------
>>> DALPath(('127.0.0.1:8000', 'public', 'profile'))
'''
StrictDALPath = typing.NewType('StrictDALPath', typing.Tuple[str, ...])
'''严格数据访问路径（只可以为字符串条目）'''
type FieldLikeType = typing.Union[str, enum.Enum, "Field"]
'''可以作为字段的类型'''
type KeyableType = typing.Union[str, int, "BaseScheme"]
"""Field value type that field as a key"""
type QueryComLikeType = typing.Union[KeyableType, "DALQueryComponent"]
'''Value type that can be used as query component'''





import typing
import enum
import datetime
import json
from ..scheme.field import FieldValueProxy
from .typing_ import JsonDumpable
from ..scheme import BaseScheme
from ..log import get_logger
from ..task.result import JsonBody

LOGGER = get_logger(__name__)


class JsonEncoder(json.JSONEncoder):
    
    def default(self, o):
        if isinstance(o, BaseScheme):
            return o.dump_to_dict()
        if isinstance(o, JsonBody):
            return o.dump_to_dict()
        if isinstance(o, FieldValueProxy):
            return o.obj
        if isinstance(o, set):
            return list(o)
        if isinstance(o, datetime.datetime):
            return o.isoformat()
        if isinstance(o, enum.Enum):
            return o.value

        return super().default(o)


def override_json_encoder(cls: type[json.JSONEncoder]):
    LOGGER.info("Json lib's encoder has been overridden")
    json.JSONEncoder = cls
    json._default_encoder = cls(
        skipkeys=False,
        ensure_ascii=True,
        check_circular=True,
        allow_nan=True,
        indent=None,
        separators=None,
        default=None,
    )

def dumps_to_json(obj: JsonDumpable) -> str:
    
    """将对象序列化为JSON字符串
    """
    return json.dumps(obj, cls=JsonEncoder)



"""Utils related to using exec.
"""

__all__ = [
    "build_func_sig"
]

import typing
from typing import Optional as Opt, Annotated as Anno, Literal as Lit


def build_func_sig(
    name: str,
    *params: tuple[str, Opt[str]],
    async_: bool = False,
    method: bool = False
) -> str:
    """
    :param params: func args, kwargs's name and annotation
    :param async_: is async function
    :param method: is method function, if True, the first param will be 'self'
    """
    params_str: str = ",".join(
        f"{i[0]}{':' + i[1] if i[1] is not None else ''}"
        for i in params
    )
    func_sig = f"{'async ' if async_ else ''}def {\
        name}({'self, ' if method else ''}{params_str}):\n"

    return func_sig



"""数据模型转换器

数据模型转换器尽可能地保证数据的类型安全、值安全，
常用于数据模型字段中
"""

import abc
import datetime
import typing
import enum
import types
from typing import Optional as Opt

from ..utils.enum_ import load_enum
from .._types import NamedTupleTV, Undefined
from .._types import _undefined
from ..utils.typing_ import JsonDumpable, get_origin, is_json_dumpable, is_namedtuple, safe_issubclass
from ..utils.main import singleton

if typing.TYPE_CHECKING:
    from . import BaseScheme


T = typing.TypeVar('T')
SchemeTV = typing.TypeVar('SchemeTV', bound='BaseScheme')
EnumMemberTV = typing.TypeVar('EnumMemberTV', bound=enum.Enum)
ConverterResultTV = typing.TypeVar('ConverterResultTV')
ConverterModeT = typing.Literal['base'] | typing.Literal['strict']
class ConverterProtocol(typing.Protocol):

    def __call__(self, value: typing.Any) -> typing.Any:
        """转换值
        """
        ...


class BaseConverter(typing.Generic[ConverterResultTV], abc.ABC):
    
    """碧霄转换器基类

    Features
    --------
    将任意输入转换为指定的类型，并保证其值合法

    模式
    ^^^^^^
    - 基本模式 `base`：即便类型不匹配，仍然尝试转换为目标类型，成功则视为通过并返回转换值
    - 严格模式 `strict`：类型不匹配则直接抛出异常
    - 子校验器不在初始化器中接受该参数，可以在调用或者直接修改mode属性来设置

    序列化
    ^^^^^^^^
    即逆转换
    
    Examples
    --------
    >>> sc = StrConverter(min_len=3, max_len=5)
    >>> sc('abc')
    'abc'
    >>> sc('ab')
    ValueError: Value 'ab' is less than minimum length 3

    """

    def __init__(self, mode: ConverterModeT = 'base') -> None:
        self.mode = mode

    @property
    def is_base(self): return self.mode == 'base'
    @property
    def is_strict(self): return self.mode == 'strict'

    @property
    @abc.abstractmethod
    def type(self) -> typing.Type[ConverterResultTV]:
        """转换结果类型注释"""
        ...
    
    @abc.abstractmethod
    def __call__(self, value, **kwargs) -> ConverterResultTV:
        """转换值
        """
        raise NotImplementedError('`__call__` method must be implemented in subclass')
    
    def dump(self, value: ConverterResultTV) -> typing.Any:

        """序列化值
        """
        return value
    
    def dump_to_str(self, value: ConverterResultTV) -> str:
        """Dump value to string
        """
        return str(value)
    
    def dump_to_jsonable(self, value: ConverterResultTV) -> JsonDumpable:
        """Dump value to jsonable value

        If not jsonable, will dump to string.
        """
        if is_json_dumpable(value):
            return value
        else:
            # raise TypeError("cannot dump type %s" % type(value))
            return self.dump_to_str(value)
    

def get_converter_from_anno(
    tp: typing.Type
) -> BaseConverter:
    
    """根据类型注释获取转换器

    Behaviour
    ---------
    - 找不到合适的校验器则返回通用校验器 :class:`AnyConverter`
    - 支持枚举、数据模型
    - 支持 NewType, UnionType, Annotated, Optional
    - Support NamedTuple

    Example
    -------
    >>> get_converter_from_anno(typing.Annotated[str, StrConverter(min_len=3)])
    StrConverter(min_len=3)
    >>> get_converter_from_anno(typing.Union[str, int])
    UnionConverter[str, int]
    >>> get_converter_from_anno(typing.Union[str, None])
    OptionalConveter[str]
    >>> get_converter_from_anno(typing.Optional[str])
    OptionalConveter[str]
    """
    from .main import BaseScheme

    ortp = get_origin(tp)
    if safe_issubclass(ortp, BaseScheme):
        return SchemeConverter(ortp)
    if safe_issubclass(ortp, enum.Enum):
        return EnumConverter(ortp)

    if ortp is int:
        return IntConverter()
    if ortp is float:
        return FloatConverter()
    if ortp is str:
        return StrConverter()
    if ortp is None:
        return NoneConverter()
    if ortp is datetime.datetime:
        return DatetimeConverter()
    if ortp is set:
        return SetConverter(
            element_type=typing.get_args(tp)[0]
        )
    if ortp is dict:
        return DictConverter(
            value_type=typing.get_args(tp)[0]
        )
    if ortp is list:
        tp_args = typing.get_args(tp)
        if len(tp_args) == 0:
            ele_tp = typing.Any
        else:
            ele_tp = tp_args[0]
        return ListConverter(
            element_type=ele_tp
        )
    if is_namedtuple(ortp):
        return NamedTupleConveter(
            namedtuple_type=tp
        )
    if ortp is tuple:
        return TupleConverter(
            tuple_type=tp
        )

    # parse union type and optional type
    if typing.get_origin(tp) is typing.Union:
        args = typing.get_args(tp)
        # if one is None, it's optional
        if len(args) == 2 and (
            args[1] is types.NoneType or args[0] is types.NoneType
        ):
            return OptionalConveter(args[1] if args[1] is not types.NoneType else args[0])

        return UnionConverter(*args)
    
    return AnyConverter()



@singleton
class AnyConverter(BaseConverter[typing.Any]):
    
    """
    通用转换器

    不做任何转换，只是返回原值
    """
    
    def __call__(self, value: typing.Any, **kwargs) -> typing.Any: 
        return value
    
    @property
    def type(self): return typing.Type[typing.Any]


class SchemeConverter(BaseConverter[SchemeTV], typing.Generic[SchemeTV]):

    """数据模型转换器
    """

    def __init__(self, 
        scheme_cls: typing.Type[SchemeTV],
        mode: ConverterModeT = 'base'
    ) -> None:
        
        super().__init__(mode)
        self.scheme_cls = scheme_cls

    def __call__(self, value: dict | SchemeTV, **kwargs) -> SchemeTV:
        """
        :param value: 序列化值
        :param kwargs: 额外参数
            - _task_context: 任务上下文
        """
        from . import BaseScheme

        if isinstance(value, dict):
            return self.scheme_cls(**value, **kwargs)
        elif isinstance(value, BaseScheme):
            for k, v in kwargs.items():
                value[k] = v
            return value
        else:
            raise ValueError('value should be dict or BaseScheme')

    @property
    def type(self): return self.scheme_cls

    def dump_to_jsonable(self, value): 
        return value.dump_to_dict(jsonable=True)


class EnumConverter(BaseConverter[EnumMemberTV], typing.Generic[EnumMemberTV]):

    """枚举转换器
    """

    def __init__(self, 
        enum_cls: typing.Type[EnumMemberTV],
        mode: ConverterModeT = 'base'
    ) -> None:
        
        super().__init__(mode)
        self.enum_cls = enum_cls

    def __call__(self, value, **kwargs) -> EnumMemberTV:
        return load_enum(self.enum_cls, value)
    
    @property
    def type(self): return self.enum_cls

    def dump_to_jsonable(self, value): return value.value
    def dump_to_str(self, value): return value.value

class EnumValueConverter(BaseConverter):

    def __call__(self, value, **kwargs) -> typing.Any:
        if not isinstance(value, enum.Enum):
            raise ValueError("not enum member")
        return value.value
    
    @property
    def type(self): return typing.Type[typing.Any]


# TODO union typevar
class UnionConverter(BaseConverter):

    """联合类型转换器

    只要一组转换器中有一个转换成功就有效（子转换器为严格模式）
    """

    def __init__(self, 
        *types_: typing.Type,
        mode: ConverterModeT = 'base'
    ) -> None:
        super().__init__(mode)

        self.sub_conveters = []
        for type_ in types_:
            converter = get_converter_from_anno(type_)
            converter.mode = 'strict'
            self.sub_conveters.append(converter)

    def __call__(self, value: typing.Any, **kwargs) -> typing.Any:

        for converter in self.sub_conveters:  # 将频次较高的放在前面，效率就更高
            try:
                return converter(value)
            except ValueError:
                continue

    @property
    def type(self): 
        return typing.Type[
            typing.Union[*tuple(validator.type for validator in self.sub_conveters)]
        ]

class OptionalConveter(BaseConverter[typing.Optional[ConverterResultTV]]):

    def __init__(self, 
        tp: Undefined | typing.Type[ConverterResultTV] = _undefined,
        tp_converter: Opt[BaseConverter[ConverterResultTV]] = None,
        mode: ConverterModeT = 'base'
    ):
        super().__init__(mode)

        if not tp_converter:
            if tp is _undefined:
                raise ValueError('`tp` or `tp_converter` must be provided')
            self.sub_converter = get_converter_from_anno(tp)
        else:
            self.sub_converter = tp_converter

    def __call__(self, value, **kwargs) -> ConverterResultTV | types.NoneType:
        
        try:
            NoneConverter()(value)
        except ValueError:
            return self.sub_converter(value)
        
    @property
    def type(self): return typing.Type[typing.Optional[self.sub_converter.type]]

    def dump_to_jsonable(self, value: ConverterResultTV) -> JsonDumpable:
        if value is None:
            return None
        else:
            return self.sub_converter.dump_to_jsonable(value)


class IntConverter(BaseConverter[int]):
    
    """整型转换器

    Features
    --------
    可以校验：
    - 大小
    - 奇偶性
    """

    def __init__(self, 
        ge: Opt[int] = None, le: Opt[int] = None,
        mode: ConverterModeT = 'base'
    ):
        
        super().__init__(mode)

        self.ge = ge
        self.le = le

    def __call__(self, value: typing.Any, **kwargs) -> int:
        
        res = int(value)
        if self.ge is not None and res <= self.ge:
            raise ValueError(f'Value {res} is less than minimum {self.ge}')
        if self.le is not None and res >= self.le:
            raise ValueError(f'Value {res} is greater than maximum {self.le}')
        return res
    
    @property
    def type(self): return int


class FloatConverter(BaseConverter[float]):
    """
    Features
    --------
    - number range
    """

    def __init__(self, 
        gt: Opt[float] = None, ge: Opt[float] = None, 
        lt: Opt[float] = None, le: Opt[float] = None,
        mode: ConverterModeT = 'base'
    ):  
        """
        If gt provided, don't provide ge, which is same for lt.
        Since ge prior to gt.
        """

        super().__init__(mode)

        self.gt = gt
        self.lt = lt
        self.ge = ge
        self.le = le

    def __call__(self, value: typing.Any, **kwargs) -> float:
        
        if not isinstance(value, float):
            if self.is_base:
                value = float(value)
            else:
                raise ValueError(f'Value {value} is not float')
        
        if self.ge is not None and value <= self.ge:
            raise ValueError(f'Value {value} is less than minimum {self.ge}')
        if self.le is not None and value >= self.le:
            raise ValueError(f'Value {value} is greater than maximum {self.le}')
        if self.gt is not None and value < self.gt:
            raise ValueError(f'Value {value} is less(or equal) than minimum {self.gt}')
        if self.lt is not None and value > self.lt:
            raise ValueError(f'Value {value} is greater(or equal) than maximum {self.lt}')
        return value
    
    @property
    def type(self): return float
    

class StrConverter(BaseConverter[str]):

    """字符串转换器
    """

    def __init__(self, 
        min: Opt[int] = None, max: Opt[int] = None,
        allow_empty: bool = False,
        half_as_unit: bool = False,
        mode: ConverterModeT = 'base'
    ):
        
        """
        :param min: 最小长度
        :param max: 最大长度
        :param allow_empty: 允许空字符串
        :param half_as_unit: 半宽字符作为长度基本单位
            
            - True: 1:1
            - False: 1:2
        """

        super().__init__(mode)

        self.min_len = min
        self.max_len = max
        self.allow_empty = allow_empty
        self.half_as_unit = half_as_unit

    def get_length(self, value: str) -> int:
        """获取字符串长度
        """
        if self.half_as_unit:
            return len(value)
        else:
            return sum(2 if ord(i) > 255 else 1 for i in value) # TESTME 

    def __call__(self, value, **kwargs) -> str:
        
        res = ''
        if self.is_base:
            res = str(value)
        if self.is_strict:
            if not isinstance(value, str):
                raise ValueError(f'Value {value} is not str')
            res = value

        length = self.get_length(res)
        if self.min_len is not None and length < self.min_len:
            if length == 0 and self.allow_empty:
                return res
            raise ValueError(f'Value {res} is less than minimum length {self.min_len}')
        if self.max_len is not None and length > self.max_len:
            raise ValueError(f'Value {res} is greater than maximum length {self.max_len}')
        
        return res
    
    @property
    def type(self): return str


@singleton
class NoneConverter(BaseConverter[None]):
    
    """None转换器

    有效值
    -----
    严格模式
    ^^^^^^^^^
    - None
    基本模式
    ^^^^^^^^^
    - 'null'
    - 'undefined'
    """

    def __call__(self, value: typing.Any, **kwargs) -> None:

        if value is not None:
            if self.is_base:
                # try from string
                if isinstance(value, str):
                    if value == 'null':
                        return None

            raise ValueError(f'Value {value} is not None')
        return None
    
    @property
    def type(self): return types.NoneType


TupleValueTV = typing.TypeVarTuple('TupleValueTV')
class TupleConverter(
    BaseConverter[tuple[typing.Unpack[TupleValueTV]]], 
    typing.Generic[typing.Unpack[TupleValueTV]]
):

    """元组转换器

    按照顺序校验每个元素的值，长度必须一致
    """

    def __init__(self, 
        tuple_type: typing.Type[tuple[typing.Unpack[TupleValueTV]]],
        mode: ConverterModeT = 'base'
    ):
        super().__init__(mode)

        self._sub_converters: tuple[BaseConverter, ...]
        tuple_type_args = typing.get_args(tuple_type)
        if len(tuple_type_args) > 1 and tuple_type_args[1] == Ellipsis:
            self._only_a_type = True
            self._sub_converters = (get_converter_from_anno(tuple_type_args[0]),)
        else:
            self._only_a_type = False
            self._sub_converters = tuple(
                get_converter_from_anno(type_)
                for type_ in tuple_type_args
            )

    def __call__(self, value, **kwargs) -> tuple[typing.Unpack[TupleValueTV]]:
        
        if not isinstance(value, tuple):
            if self.is_base:
                value = tuple(value)

        if not self._only_a_type:
            if len(value) != len(self._sub_converters):
                raise ValueError(f'Value {value} is not a tuple of length {len(self._sub_converters)}')
        
            return tuple(
                validator(value[i])
                for i, validator in enumerate(self._sub_converters)
            )
        else:
            return tuple(
                self._sub_converters[0](i)
                for i in value
            )
    
    @property
    def type(self): return typing.Tuple[*tuple(
        validator.type 
        for validator in self._sub_converters
    )]

    def dump_to_jsonable(self, value) -> tuple:
        if not self._only_a_type:
            return tuple(
                self._sub_converters[i].dump_to_jsonable(value[i])
                for i in range(len(value))
            )
        else:
            return tuple(
                self._sub_converters[0].dump_to_jsonable(i)
                for i in value
            )
    

class NamedTupleConveter(BaseConverter[NamedTupleTV], typing.Generic[NamedTupleTV]):
    """带名元组转换器
    """

    def __init__(self, 
        namedtuple_type: typing.Type[NamedTupleTV],
        mode: ConverterModeT = 'base'
    ):
        super().__init__(mode)
        self.namedtuple_cls: typing.Type[NamedTupleTV] = namedtuple_type
        self.sub_validators = tuple(
            get_converter_from_anno(i)
            for i in self.namedtuple_cls.__annotations__.values()
        )
    
    def __call__(self, value, **kwargs) -> NamedTupleTV:

        if type(value) is self.namedtuple_cls:
            return value
        
        if not isinstance(value, tuple):
            value = TupleConverter(typing.Tuple[typing.Any, ...])(value)
        
        if len(value) != len(self.namedtuple_cls._fields):
            raise ValueError(f"Value {value} is not a namedtuple of length {len(self.namedtuple_cls._fields)}")
        
        # TODO support default value
        
        converted_values = tuple(
            self.sub_validators[i](value[i])
            for i in range(len(value))
        )
        return self.namedtuple_cls(*converted_values) # type: ignore
        # FIXME type issue
    
    @property
    def type(self): return self.namedtuple_cls


class SetConverter(BaseConverter[typing.Set[T]], typing.Generic[T]):
    """集合转换器
    
    Behaviour
    ---------
    校验内容
    ^^^^^^^^^
    - 是集合
    - 集合元素类型
    """

    def __init__(self, 
        element_type: typing.Type[T],
        mode: ConverterModeT = 'base'
    ):
        
        super().__init__(mode)
        self.sub_conveter: BaseConverter[T] = get_converter_from_anno(element_type)

    def __call__(self, value, **kwargs) -> set:
        # is a set
        if not isinstance(value, set):
            if self.is_base:
                value = set(value)
            else:
                raise ValueError(f"Value {value} is not a set")
        
        # validate element type
        new_value: typing.Set[T] = set()
        for i in value:
            new_value.add(self.sub_conveter(i))
        
        return new_value
    
    def dump(self, value: set) -> tuple:
        return tuple(
            i
            for i in value
        )
    
    def dump_to_jsonable(self, value) -> set: 
        return set(
            self.sub_conveter.dump_to_jsonable(i)
            for i in value
        )
    
    @property
    def type(self): return set


class ListConverter(BaseConverter[typing.List[T]], typing.Generic[T]):

    """列表转换器

    校验内容
    ^^^^^^^^^
    - 是列表
    - 列表元素类型
    - 列表长度
    
    TODO
    - 元素不重复
        - 如果不是简单元素，需要指定标识符获取器
    """

    def __init__(self, 
        element_type: typing.Type[T],
        min_len: Opt[int] = None, max_len: Opt[int] = None,
        mode: ConverterModeT = 'base'
    ):
        super().__init__(mode)

        self.min_len = min_len
        self.max_len = max_len
        converter = get_converter_from_anno(element_type)
        converter.mode = mode
        self.sub_converter: BaseConverter[T] = converter

    def __call__(self, value, **kwargs) -> typing.List[T]:

        # is a list
        if not isinstance(value, list):
            if self.is_base:
                value = list(value)
            
            raise ValueError(f"Value {value} is not list")
        
        # validate length
        length = len(value)
        if self.min_len is not None and length < self.min_len:
            # TODO use (Container)LengthError
            raise ValueError(f'Value {value} is less than minimum length {self.min_len}')
        if self.max_len is not None and length > self.max_len:
            raise ValueError(f'Value {value} is greater than maximum length {self.max_len}')

        # validate element type
        new_value: typing.List[T] = []
        for i in value:
            new_value.append(self.sub_converter(i))
        
        return new_value
    
    @property
    def type(self): return typing.List[self.sub_converter.type]

    def dump_to_jsonable(self, value) -> list:
        return [
            self.sub_converter.dump_to_jsonable(i)
            for i in value
        ]
        

class DatetimeConverter(BaseConverter[datetime.datetime]):
    """日期时间转换器

    是否为 datetime.datetime 对象
    """

    def __call__(self, value, **kwargs) -> datetime.datetime:
        
        if not isinstance(value, datetime.datetime):
            if self.is_base:
                if isinstance(value, str):
                    return datetime.datetime.fromisoformat(value)
                if isinstance(value, (int, float)):
                    return datetime.datetime.fromtimestamp(value)
                
            raise ValueError(f"Value {value} is not datetime.datetime obj")
        
        return value
    
    def dump(self, value) -> str:
        return value.isoformat()

    def dump_to_jsonable(self, value):
        return value.isoformat()
    
    @property
    def type(self): return datetime.datetime


class TimeConverter(BaseConverter[datetime.time]):

    """时间校验器

    是否为 datetime.time 对象
    """

    def __call__(self, value, **kwargs) -> datetime.time:
        
        if not isinstance(value, datetime.time):
            if self.is_base:
                if isinstance(value, str):
                    return datetime.time.fromisoformat(value)
                
            raise ValueError(f"Value {value} is not datetime.time obj")
        
        return value
    
    def dump(self, value):
        """序列化为 ISO 格式的字符串
        """
        return value.isoformat()
    
    @property
    def type(self): return datetime.time

    def dump_to_jsonable(self, value): 
        return value.isoformat()


class DictConverter(BaseConverter[typing.Dict[typing.Any, T]], typing.Generic[T]):

    """字典校验器

    校验内容
    ^^^^^^^^^
    - 是字典
    - 字典值类型
    """

    def __init__(self, 
        value_type: typing.Type[T],
        mode: ConverterModeT = 'base'
    ):
        
        super().__init__(mode)
        self.value_conveter = get_converter_from_anno(value_type)

    def __call__(self, value, **kwargs) -> typing.Dict[typing.Any, T]:

        # is a dict
        if not isinstance(value, dict):
            if self.is_base:
                value = dict(value)
            
            raise ValueError(f"Value {value} is not dict")
        
        # validate element type
        new_value: typing.Dict[typing.Any, T] = {}
        for k, v in value.items():
            new_value[k] = self.value_conveter(v)
        
        return new_value
    
    @property
    def type(self): 
        return typing.Dict[typing.Any, self.value_conveter.type]
    
    def dump_to_jsonable(self, value) -> dict[str, JsonDumpable]: 
        return {
            str(k): self.value_conveter.dump_to_jsonable(v)
            for k, v in value.items()
        }
